# -*- coding: utf-8 -*-
#
# Copyright 2011 Manuel Stocker <mensi@mensi.ch>
#
# This file is part of GitServerGlue.
#
# GitServerGlue is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GitServerGlue is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GitServerGlue.  If not, see http://www.gnu.org/licenses

from twisted.cred import portal
from twisted.conch import avatar
from twisted.conch.checkers import SSHPublicKeyDatabase
from twisted.conch.ssh import factory, session
from twisted.internet import reactor, defer
from twisted.internet.error import ProcessExitedAlready
from twisted.python import log, components
from zope.interface import implements
import shlex

from gitserverglue.common import ErrorProcess, PasswordChecker


class GitAvatar(avatar.ConchUser):
    def __init__(self, username, authnz, git_configuration):
        avatar.ConchUser.__init__(self)
        self.username = username
        self.authnz = authnz
        self.git_configuration = git_configuration
        self.channelLookup.update({'session': session.SSHSession})


class GitRealm:
    implements(portal.IRealm)

    def __init__(self, authnz, git_configuration):
        self.authnz = authnz
        self.git_configuration = git_configuration

    def requestAvatar(self, avatarId, mind, *interfaces):
        log.msg("request %r, %r" % (mind, interfaces))
        return interfaces[0], GitAvatar(avatarId, self.authnz,
                                        self.git_configuration), lambda: None


class GitSession:
    def __init__(self, avatar):
        self.avatar = avatar
        self.ptrans = None

    def execCommand(self, proto, cmd):
        cmdparts = shlex.split(cmd)
        rpc = cmdparts[0]
        path = cmdparts[-1]

        path_info = self.avatar.git_configuration.path_lookup(path,
                                    protocol_hint="ssh")
        if path_info is None or path_info['repository_fs_path'] is None:
            log.msg('User %s tried to access %s but the translator did '
                    'not return a real path' % (self.avatar.username, path))
            return self._kill_connection(proto, "Unknown Repository")

        if rpc not in ['git-upload-pack', 'git-receive-pack']:
            log.err('Unknown RPC: ' + rpc)
            return self._kill_connection(proto, "Unknown RPC")

        if (rpc == 'git-upload-pack' and
            not self.avatar.authnz.can_read(self.avatar.username, path_info)):
            log.msg('User %s tried to access %s but '
                    'does not have read permissions' % (self.avatar.username,
                                                            path))
            return self._kill_connection(proto,
                                         "You don't have read permissions")

        if (rpc == 'git-receive-pack' and
            not self.avatar.authnz.can_write(self.avatar.username, path_info)):
            log.msg('User %s tried to access %s but does not have '
                    'write permissions' % (self.avatar.username, path))
            return self._kill_connection(proto,
                                         "You don't have write permissions")

        gitshell = self.avatar.git_configuration.git_shell_binary
        cmdargs = ['git-shell', '-c',
                   rpc + ' \'' + path_info['repository_fs_path'] + '\'']
        log.msg("Spawning %s with args %r" % (gitshell, cmdargs))
        self.ptrans = reactor.spawnProcess(proto, gitshell, cmdargs)

    def getPty(self, term, windowSize, attrs):
        pass

    def openShell(self, proto):
        self._kill_connection(proto, "Shell access not allowed\n")

    def _kill_connection(self, proto, msg):
        ErrorProcess(proto, 128, msg)

    def eofReceived(self):
        if self.ptrans:
            self.ptrans.closeStdin()

    def closed(self):
        if self.ptrans:
            try:
                self.ptrans.signalProcess('HUP')
            except (OSError, ProcessExitedAlready):
                pass
            self.ptrans.loseConnection()


components.registerAdapter(GitSession, GitAvatar, session.ISession)


class PublicKeyChecker(SSHPublicKeyDatabase):
    def __init__(self, checker):
        self.checker = checker

    def checkKey(self, credentials):
        return defer.maybeDeferred(self.checker, credentials.username,
                                   credentials.blob)


def create_factory(private_keys, public_keys, authnz, git_configuration):
    class GitSSHFactory(factory.SSHFactory):
        publicKeys = public_keys
        privateKeys = private_keys

    gitportal = portal.Portal(GitRealm(authnz, git_configuration))

    if hasattr(authnz, 'check_password'):
        log.msg("Registering PasswordChecker")
        gitportal.registerChecker(PasswordChecker(authnz.check_password))
    if hasattr(authnz, 'check_publickey'):
        log.msg("Registering PublicKeyChecker")
        gitportal.registerChecker(PublicKeyChecker(authnz.check_publickey))

    GitSSHFactory.portal = gitportal

    return GitSSHFactory()



# -*- coding: utf-8 -*-
#
# Copyright 2011 Manuel Stocker <mensi@mensi.ch>
#
# This file is part of GitServerGlue.
#
# GitServerGlue is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GitServerGlue is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GitServerGlue.  If not, see http://www.gnu.org/licenses

import os.path
import re
import datetime
import calendar
import email.utils

from zope.interface import implements
from twisted.python import log

from twisted.internet import reactor, defer, task
from twisted.internet.interfaces import IProcessProtocol
from twisted.internet.interfaces import IPushProducer, IConsumer

from twisted.cred.portal import IRealm, Portal
from twisted.cred.checkers import AllowAnonymousAccess, ANONYMOUS
from twisted.web.guard import HTTPAuthSessionWrapper, BasicCredentialFactory
from twisted.web._auth.wrapper import UnauthorizedResource

from twisted.web.static import File
from twisted.web.server import Site, NOT_DONE_YET
from twisted.web.resource import Resource, IResource
from twisted.web.resource import NoResource, ForbiddenResource

from gitserverglue.common import PasswordChecker, git_packet
from gitserverglue.streamingweb import StreamingRequest


def get_date_header(dt=None):
    """Get the HTTP header value for a given datetime"""
    if dt is None:
        dt = datetime.datetime.now()
    t = calendar.timegm(dt.utctimetuple())
    return email.utils.formatdate(t, localtime=False, usegmt=True)

cache_forever = lambda: [('Expires', get_date_header(datetime.datetime.now() +
                                                datetime.timedelta(days=365))),
                 ('Pragma', 'no-cache'),
                 ('Cache-Control', 'public, max-age=31556926')]

dont_cache = lambda: [('Expires', 'Fri, 01 Jan 1980 00:00:00 GMT'),
              ('Pragma', 'no-cache'),
              ('Cache-Control', 'no-cache, max-age=0, must-revalidate')]

file_headers = {
    re.compile('.*/(HEAD)$'):                                   lambda: dict(dont_cache() + [('Content-Type', 'text/plain')]),
    re.compile('.*/(objects/info/alternates)$'):                lambda: dict(dont_cache() + [('Content-Type', 'text/plain')]),
    re.compile('.*/(objects/info/http-alternates)$'):           lambda: dict(dont_cache() + [('Content-Type', 'text/plain')]),
    re.compile('.*/(objects/info/packs)$'):                     lambda: dict(dont_cache() + [('Content-Type', 'text/plain; charset=utf-8')]),
    re.compile('.*/(objects/info/[^/]+)$'):                     lambda: dict(dont_cache() + [('Content-Type', 'text/plain')]),
    re.compile('.*/(objects/[0-9a-f]{2}/[0-9a-f]{38})$'):       lambda: dict(cache_forever() + [('Content-Type', 'application/x-git-loose-object')]),
    re.compile('.*/(objects/pack/pack-[0-9a-f]{40}\\.pack)$'):  lambda: dict(cache_forever() + [('Content-Type', 'application/x-git-packed-objects')]),
    re.compile('.*/(objects/pack/pack-[0-9a-f]{40}\\.idx)$'):   lambda: dict(cache_forever() + [('Content-Type', 'application/x-git-packed-objects-toc')]),
}


class FileLikeProducer(object):
    """twisted.web.client.FileBodyProducer adaptation for request.content"""
    implements(IPushProducer)

    def __init__(self, inputFile, consumer=None, cooperator=task,
                 readSize=2 ** 16):
        self._inputFile = inputFile
        self._cooperate = cooperator.cooperate
        self._readSize = readSize
        self._consumer = consumer
        self._task = None

    def stopProducing(self):
        try:
            self._task.stop()
        except:
            pass

    def startProducing(self, consumer=None):
        if consumer is not None:
            self._consumer = consumer

        if self._task is not None:
            raise Exception("Already started")

        self._createTask()

    def pauseProducing(self):
        self._task.pause()

    def resumeProducing(self):
        if self._task is None:
            self._createTask()
        self._task.resume()

    def _createTask(self):
        if self._consumer is None:
            raise Exception("No consumer")

        self._task = self._cooperate(self._writeloop(self._consumer))
        d = self._task.whenDone()

        def maybeStopped(reason):
            reason.trap(task.TaskStopped)
            return defer.Deferred()

        d.addCallbacks(lambda ignored: None, maybeStopped)
        return d

    def _writeloop(self, consumer):
        while True:
            data = self._inputFile.read(self._readSize)
            if not data:
                break
            consumer.write(data)
            yield None


class GitCommand(Resource):
    """A resource returning content from a git process"""
    implements(IProcessProtocol, IConsumer)

    isLeaf = True
    process = None
    _producer = None

    def __init__(self, cmd, args):
        self.cmd = cmd
        self.args = args

    # Resource
    def render(self, request):
        self.request = request
        reactor.spawnProcess(self, self.cmd, self.args)

        return NOT_DONE_YET

    # IProcessProtocol
    def makeConnection(self, process):
        self.process = process

        # twisted.internet.process.Process seems to not fully
        # implement IPushProducer since stopProducing is missing
        # therefore patch in a dummy one
        if not hasattr(process, "stopProducing"):
            setattr(process, "stopProducing",
                    lambda: process.loseConnection())

        self.request.registerProducer(process, True)

        if not isinstance(self.request, StreamingRequest):
            # twisted default request, does not support streaming contents
            producer = FileLikeProducer(self.request.content, process)
            producer.startProducing()
            process.registerProducer(producer, True)

        elif self._producer is not None:
            self._producer.resumeProducing()
            self.process.registerProducer(self._producer, True)

    def childDataReceived(self, childFD, data):
        self.request.write(data)

    def childConnectionLost(self, childFD):
        pass

    def processExited(self, reason):
        pass

    def processEnded(self, reason):
        self.request.unregisterProducer()
        self.request.finish()

    # IConsumer for StreamingRequest
    def registerProducer(self, producer, streaming):
        # suppress original stopProducing since git will
        # call it on closing stdin -> httpchannel gets closed
        # because the original stopProducing calls loseConnection
        def suppressedStop(self):
            # Do not pause producing here unless you make sure
            # producing is resumed upon request completion. Otherwise,
            # no further requests will be received in keep-alive!
            pass  # self.pauseProducing()
        suppressedStop.original = producer.stopProducing
        producer.stopProducing = suppressedStop.__get__(producer,
                                                        producer.__class__)

        if self.process is None:
            producer.pauseProducing()
            self._producer = producer
        else:
            self.process.registerProducer(producer)

    def unregisterProducer(self):
        if self.process:
            self.process.closeStdin()

    def write(self, data):
        self.process.write(data)


class InfoRefs(Resource):
    """Resource for handling git requests to /info/refs"""
    isLeaf = True

    def __init__(self, gitpath, gitcommand='git'):
        self.gitpath = gitpath
        self.gitcommand = gitcommand

    def render_GET(self, request):
        if 'service' not in request.args:
            # dumb client
            for key, val in dont_cache():
                request.setHeader(key, val)

            log.msg('Dumb client, sending %s' % os.path.join(
                                                self.gitpath, 'info', 'refs'))
            return File(os.path.join(self.gitpath, 'info', 'refs'),
                        'text/plain; charset=utf-8').render_GET(request)

        else:
            # smart client
            if request.args['service'][0] not in ['git-upload-pack',
                                                  'git-receive-pack']:
                return "Invalid RPC: " + request.args['service'][0]

            rpc = request.args['service'][0][4:]
            request.write(git_packet('# service=git-' + rpc) + git_packet())
            cmd = self.gitcommand
            args = [os.path.basename(cmd), rpc, '--stateless-rpc',
                    '--advertise-refs', self.gitpath]

            return GitCommand(cmd, args).render(request)


class GitResource(Resource):
    """Resource representing a git repository"""

    def __init__(self, username, authnz, git_configuration,
                 credentialFactories, git_viewer):
        Resource.__init__(self)

        self.username = username
        self.authnz = authnz
        self.git_configuration = git_configuration
        self.credentialFactories = credentialFactories
        self.git_viewer = git_viewer

    def getChild(self, path, request):
        """Find the appropriate child resource depending on request type

        Possible URLs:
        - /foo/bar/info/refs -> info refs (file / SmartHTTP hybrid)
        - /foo/bar/git-upload-pack -> SmartHTTP RPC
        - /foo/bar/git-receive-pack -> SmartHTTP RPC
        - /foo/bar/HEAD -> file (dumb http)
        - /foo/bar/objects/* -> file (dumb http)
        """
        path = request.path  # alternatively use path + request.postpath
        pathparts = path.split('/')
        writerequired = False
        script_name = '/'
        new_path = path
        resource = NoResource()

        # Path lookup / translation
        path_info = self.git_configuration.path_lookup(path,
                                                       protocol_hint='http')
        if path_info is None:
            log.msg('User %s tried to access %s '
                    'but the lookup failed' % (self.username, path))
            return resource

        log.msg('Lookup of %s gave %r' % (path, path_info))

        if (path_info['repository_fs_path'] is None and
            path_info['repository_base_fs_path'] is None):
            log.msg('Neither a repository base nor a repository were returned')
            return resource

        # split script_name / new_path according to path info
        if path_info['repository_base_url_path'] is not None:
            script_name = '/'
            script_name += path_info['repository_base_url_path'].strip('/')
            new_path = path[len(script_name.rstrip('/')):]

        # since pretty much everything needs read access, check for it now
        if not self.authnz.can_read(self.username, path_info):
            if self.username is None:
                return UnauthorizedResource(self.credentialFactories)
            else:
                return ForbiddenResource("You don't have read access")

        # Smart HTTP requests
        # /info/refs
        if (len(pathparts) >= 2 and
            pathparts[-2] == 'info' and
            pathparts[-1] == 'refs'):
            writerequired = ('service' in request.args and
                             request.args['service'][0] == 'git-receive-pack')
            resource = InfoRefs(path_info['repository_fs_path'])

        # /git-upload-pack (client pull)
        elif len(pathparts) >= 1 and pathparts[-1] == 'git-upload-pack':
            cmd = 'git'
            args = [os.path.basename(cmd), 'upload-pack', '--stateless-rpc',
                    path_info['repository_fs_path']]
            resource = GitCommand(cmd, args)
            request.setHeader('Content-Type',
                              'application/x-git-upload-pack-result')

        # /git-receive-pack (client push)
        elif len(pathparts) >= 1 and pathparts[-1] == 'git-receive-pack':
            writerequired = True
            cmd = 'git'
            args = [os.path.basename(cmd), 'receive-pack',
                    '--stateless-rpc', path_info['repository_fs_path']]
            resource = GitCommand(cmd, args)
            request.setHeader('Content-Type',
                              'application/x-git-receive-pack-result')

        # static files as specified in file_headers or fallback webfrontend
        else:
            # determine the headers for this file
            filename, headers = None, None
            for matcher, get_headers in file_headers.items():
                m = matcher.match(path)
                if m:
                    filename = m.group(1)
                    headers = get_headers()
                    break

            if filename is not None:
                for key, val in headers.items():
                    request.setHeader(key, val)

                log.msg("Returning file %s" % os.path.join(
                                    path_info['repository_fs_path'], filename))
                resource = File(os.path.join(path_info['repository_fs_path'],
                                        filename), headers['Content-Type'])
                resource.isLeaf = True  # static file -> it is a leaf

            else:
                # No match -> fallback to git viewer
                if script_name is not None:
                    # patch pre/post path of request according to
                    # script_name and path
                    request.prepath = script_name.strip('/').split('/')
                    if request.prepath == ['']:
                        request.prepath = []
                    request.postpath = new_path.lstrip('/').split('/')

                    log.msg("pre and post: %r %r" % (request.prepath,
                                                     request.postpath))

                # If the resource has a withEnviron function, it's
                # probably our own flavour of WSGIResource that
                # supports passing further args for the environ
                if hasattr(self.git_viewer, "withEnviron"):
                    # set wsgirouting args
                    routing_args = {
                        'repository_path': path_info['repository_fs_path'],
                        'repository_base': path_info['repository_base_fs_path']
                    }
                    if 'repository_clone_urls' in path_info:
                        routing_args['repository_clone_urls'] = path_info['repository_clone_urls']
                    resource = self.git_viewer.withEnviron(
                                    {'wsgiorg.routing_args': ([], routing_args)})
                else:
                    resource = self.git_viewer

        # before returning the resource, check if write access is required
        # and enforce privileges accordingly
        # anonymous (username = None) will never be granted write access
        if writerequired and (self.username is None or
                              not self.authnz.can_write(self.username,
                                                        path_info)):
            if self.username is None:
                return UnauthorizedResource(self.credentialFactories)
            else:
                return ForbiddenResource("You don't have write access")

        return resource

    def render_GET(self, request):
        return ForbiddenResource()


class GitHTMLRealm(object):
    implements(IRealm)

    def __init__(self, authnz, git_configuration,
                 credentialFactories, git_viewer):
        self.authnz = authnz
        self.git_configuration = git_configuration
        self.credentialFactories = credentialFactories
        self.git_viewer = git_viewer

    def requestAvatar(self, avatarId, mind, *interfaces):
        if avatarId == ANONYMOUS:
            avatarId = None  # anonymous

        if IResource in interfaces:
            return IResource, GitResource(avatarId, self.authnz,
                                          self.git_configuration,
                                          self.credentialFactories,
                                          self.git_viewer), lambda: None
        raise NotImplementedError()


def create_factory(authnz, git_configuration, git_viewer=None):
    if git_viewer is None:
        git_viewer = NoResource()
    elif not IResource.providedBy(git_viewer):
        raise ValueError("git_viewer should be either implement IResource")

    credentialFactories = [BasicCredentialFactory('Git Repositories')]
    gitportal = Portal(GitHTMLRealm(authnz, git_configuration,
                                    credentialFactories, git_viewer))

    if hasattr(authnz, 'check_password'):
        log.msg("Registering PasswordChecker")
        gitportal.registerChecker(PasswordChecker(authnz.check_password))
    gitportal.registerChecker(AllowAnonymousAccess())

    resource = HTTPAuthSessionWrapper(gitportal, credentialFactories)
    site = Site(resource)

    return site



# -*- coding: utf-8 -*-
#
# Copyright 2011 Manuel Stocker <mensi@mensi.ch>
#
# This file is part of GitServerGlue.
#
# GitServerGlue is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GitServerGlue is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GitServerGlue.  If not, see http://www.gnu.org/licenses

import os
import os.path
import sys

from twisted.internet import reactor
from twisted.conch.ssh import keys
from twisted.python import log

from Crypto.PublicKey import RSA

from gitserverglue import ssh, http, git
from gitserverglue.streamingweb import make_site_streaming
from gitserverglue.wsgihelper import WSGIResource

from ConfigParser import SafeConfigParser
from passlib.apache import HtpasswdFile


class TestAuthnz(object):
    def __init__(self,
                 htpasswd_file=".htpasswd",
                 perms_file=".repoperms",
                 keys_file=".rsakeys"):
        self.htpasswd = HtpasswdFile(htpasswd_file)
        self.perms_file = perms_file
        self.keys_file = keys_file

    def can_read(self, username, path_info):
        return self._check_access(username, path_info, "r")

    def can_write(self, username, path_info):
        return self._check_access(username, path_info, "w")

    def _check_access(self, username, path_info, level):
        if username is None:
            username = "anonymous"

        if path_info['repository_fs_path'] is None:
            return False

        repo = os.path.basename(path_info['repository_fs_path'])

        config = SafeConfigParser()
        config.read(self.perms_file)

        if not config.has_option(repo, username):
            return False

        return level in config.get(repo, username)

    def check_password(self, username, password):
        self.htpasswd.load_if_changed()
        return self.htpasswd.check_password(username, password)

    def check_publickey(self, username, keyblob):
        with open(self.keys_file, 'rb') as f:
            for line in f:
                try:
                    user, key = line.split(':', 1)
                    if (username == user.strip() and
                        keyblob == keys.Key.fromString(data=key.strip()
                                                       ).blob()):
                        return True
                except:
                    log.err(None, "Loading key failed")
        return False


class TestGitConfiguration(object):
    git_binary = 'git'
    git_shell_binary = 'git-shell'

    def path_lookup(self, url, protocol_hint=None):
        res = {
            'repository_base_fs_path': './',
            'repository_base_url_path': '/',
            'repository_fs_path': None
        }

        pathparts = url.strip('/').split('/')

        if len(pathparts) > 0 and pathparts[0].endswith('.git'):
            p = os.path.join('./', pathparts[0])
            if os.path.exists(p):
                res['repository_fs_path'] = p
                res['repository_clone_urls'] = {
                    'http': 'http://localhost:8080/' + pathparts[0],
                    'git': 'git://localhost/' + pathparts[0],
                    'ssh': 'ssh://localhost:5522/' + pathparts[0]
                }

        return res


def find_git_viewer():
    """Tries to find a known git viewer"""
    # pyggi - https://www.0xdeadbeef.ch/pyggi/pyggi.git/
    # not to be confused with PyGGI from PyPI
    try:
        from pyggi.lib.config import config

        config.add_section('general')
        config.set('general', 'preserve_daemon_export', "false")
        config.set('general', 'name', "pyggi")

        config.add_section('clone')

        config.add_section('modules')
        config.set('modules', 'pyggi.repositories.frontend', '/')
        config.set('modules', 'pyggi.base.base', '/')

        from pyggi import create_app

        return WSGIResource(reactor, reactor.getThreadPool(), create_app())

    except:
        pass


def main():
    log.startLogging(sys.stderr)

    keylocation = os.path.expanduser(
                    os.path.join('~', '.gitserverglue', 'key.pem'))
    key = None

    if os.path.exists(keylocation):
        try:
            key = keys.Key.fromFile(keylocation)
        except:
            pass

    if key is None:
        rsakey = RSA.generate(2048)
        key = keys.Key(rsakey)

        try:
            if not os.path.exists(os.path.dirname(keylocation)):
                os.mkdir(os.path.dirname(keylocation))
            f = open(keylocation, 'wb')
            f.write(rsakey.exportKey('PEM'))
            f.close()
        except:
            log.err(None, "Failed to write key to " + keylocation)

    ssh_factory = ssh.create_factory(
        public_keys={'ssh-rsa': key},
        private_keys={'ssh-rsa': key},
        authnz=TestAuthnz(),
        git_configuration=TestGitConfiguration()
    )

    http_factory = http.create_factory(
        authnz=TestAuthnz(),
        git_configuration=TestGitConfiguration(),
        git_viewer=find_git_viewer()
    )

    git_factory = git.create_factory(
        authnz=TestAuthnz(),
        git_configuration=TestGitConfiguration()
    )

    reactor.listenTCP(5522, ssh_factory)
    reactor.listenTCP(8080, make_site_streaming(http_factory))
    reactor.listenTCP(9418, git_factory)
    reactor.run()



# -*- coding: utf-8 -*-
#
# Copyright 2011 Manuel Stocker <mensi@mensi.ch>
#
# This file is part of GitServerGlue.
#
# GitServerGlue is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GitServerGlue is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GitServerGlue.  If not, see http://www.gnu.org/licenses

from urllib import unquote
import string

from twisted.internet.interfaces import IConsumer
from twisted.web.http import HTTPChannel, parse_qs, datetimeToString
from twisted.web.server import Request, version
from twisted.web.resource import IResource, getChildForRequest
from twisted.web.util import DeferredResource
from twisted.python import failure


class StreamingRequest(Request):
    """Modified Request to support streaming content"""

    _fallbackToBuffered = True
    fallbackContentTypes = [
        'multipart/form-data',
        'application/x-www-form-urlencoded'
    ]

    def requestHeadersReceived(self, command, path, version):
        """Called when a StreamingHTTPChannel received all headers

        If the method was POST and we have a content type listed
        in self.fallbackContentTypes, fallback mode will stay
        enabled and everything works just like in vanilla twisted.

        However, if these conditions are not met, streaming
        mode will be used."""

        ctype = self.requestHeaders.getRawHeaders('content-type')
        if ctype is not None:
            ctype = ctype[0]
        if command == "POST" and ctype in self.fallbackContentTypes:
            return  # stay in fallback mode

        if command not in ["POST", "PUT"]:
            return  # stay in fallback mode except for POST and PUT

        self._fallbackToBuffered = False

        # the following code will perform the same steps as
        # the vanilla requestReceived minus content parsing
        self.args = {}
        self.stack = []

        self.method, self.uri = command, path
        self.clientproto = version
        x = self.uri.split('?', 1)

        if len(x) == 1:
            self.path = self.uri
        else:
            self.path, argstring = x
            self.args = parse_qs(argstring, 1)

        # cache the client and server information, we'll need this later to be
        # serialized and sent with the request so CGIs will work remotely
        self.client = self.channel.transport.getPeer()
        self.host = self.channel.transport.getHost()

        self.process()

    def requestReceived(self, command, path, version):
        if self._fallbackToBuffered:
            Request.requestReceived(self, command, path, version)
        else:
            self.resource.unregisterProducer()

    def process(self):
        if self._fallbackToBuffered:
            return Request.process(self)

        # streaming mode
        # pause producing on channel until we know what resource
        # we deal with
        self.channel.transport.pauseProducing()
        self.site = self.channel.site

        self.setHeader('server', version)
        self.setHeader('date', datetimeToString())

        self.prepath = []
        self.postpath = map(unquote, string.split(self.path[1:], '/'))

        try:
            self.processResource(self.site.getResourceFor(self))
        except:
            self.processingFailed(failure.Failure())

    def processResource(self, resource):
        if not isinstance(resource, DeferredResource):
            resource = getChildForRequest(resource, self)

        if isinstance(resource, DeferredResource):
            resource.d.addCallback(self.processResource).addErrback(
                self.processingFailed)
        else:
            self.resource = self._getRealResource(resource)

            if IConsumer.providedBy(self.resource):
                self.resource.registerProducer(self.channel.transport, True)
                self.render(resource)  # use resource here to not break proxies
            else:
                self._fallbackToBuffered = True

            # resource is determined, resume producing
            self.channel.transport.resumeProducing()

    def _getRealResource(self, resource):
        proxied = False
        for b in resource.__class__.__bases__:
            if b.__name__ == '(Proxy for twisted.web.resource.IResource)':
                proxied = True

        if not proxied:
            return resource

        # find the real resource (or the next proxy)
        for x in vars(resource).values():
            if IResource.providedBy(x):
                return self._getRealResource(x)

        raise Exception("Unable to find real resource")

    def handleContentChunk(self, data):
        if self._fallbackToBuffered:
            self.content.write(data)
        else:
            self.resource.write(data)


class StreamingHTTPChannel(HTTPChannel):
    """Modified HTTPChannel to support streaming requests"""

    def allHeadersReceived(self):
        HTTPChannel.allHeadersReceived(self)
        req = self.requests[-1]
        if hasattr(req, "requestHeadersReceived"):
            req.requestHeadersReceived(self._command,
                                       self._path, self._version)


def make_site_streaming(site):
    site.requestFactory = StreamingRequest
    site.protocol = StreamingHTTPChannel

    return site



# -*- coding: utf-8 -*-
#
# Copyright 2011 Manuel Stocker <mensi@mensi.ch>
#
# This file is part of GitServerGlue.
#
# GitServerGlue is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GitServerGlue is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GitServerGlue.  If not, see http://www.gnu.org/licenses

from twisted.cred import checkers, credentials, error
from twisted.internet import defer
from twisted.internet.error import ProcessTerminated
from twisted.internet.interfaces import IProcessTransport
from twisted.python.failure import Failure
from zope.interface import implements


def git_packet(data=None):
    """Format data as a git pktline

    If data is None, a flush packet will be returned"""
    if data is None:
        return '0000'
    return str(hex(len(data) + 4)[2:].rjust(4, '0')) + data


class ErrorProcess(object):
    """Simulates a process transport with a message on stderr

    Implements an IProcessTransport that simulates process
    termination with the given return code and writing
    a message to stderr. Properly closes child FDs.
    """
    implements(IProcessTransport)

    def __init__(self, proto, code, message):
        proto.makeConnection(self)
        proto.childDataReceived(2, message + '\n')

        proto.childConnectionLost(0)
        proto.childConnectionLost(1)
        proto.childConnectionLost(2)

        failure = Failure(ProcessTerminated(code))

        proto.processExited(failure)
        proto.processEnded(failure)

        # ignore all unused methods
        noop = lambda *args, **kwargs: None
        self.closeStdin = noop
        self.closeStdout = noop
        self.closeStderr = noop
        self.writeToChild = noop
        self.loseConnection = noop
        self.signalProcess = noop


class PasswordChecker:

    implements(checkers.ICredentialsChecker)
    credentialInterfaces = (credentials.IUsernamePassword,)

    def __init__(self, checker):
        self.checker = checker

    def _cbPasswordMatch(self, matched, username):
        if matched:
            return defer.succeed(username)
        else:
            return Failure(error.UnauthorizedLogin())

    def requestAvatarId(self, credentials):
        return defer.maybeDeferred(
           self.checker,
           credentials.username,
           credentials.password
        ).addCallback(self._cbPasswordMatch, str(credentials.username))



# -*- coding: utf-8 -*-
#
# Copyright 2011 Manuel Stocker <mensi@mensi.ch>
#
# This file is part of GitServerGlue.
#
# GitServerGlue is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GitServerGlue is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GitServerGlue.  If not, see http://www.gnu.org/licenses

from twisted.web.server import NOT_DONE_YET
import twisted.web.wsgi as twsgi


class _WSGIResponse(twsgi._WSGIResponse):
    def __init__(self, reactor, threadpool, application, request,
                 environ=None):
        twsgi._WSGIResponse.__init__(self, reactor, threadpool, application,
                                     request)

        if environ is not None:
            self.environ.update(environ)


class WSGIResource(twsgi.WSGIResource):
    def __init__(self, reactor, threadpool, application, environ=None):
        twsgi.WSGIResource.__init__(self, reactor, threadpool, application)

        self.environ = environ

    def render(self, request):
        response = _WSGIResponse(
            self._reactor, self._threadpool,
            self._application, request, self.environ)
        response.start()
        return NOT_DONE_YET

    def withEnviron(self, environ):
        """Create a new WSGIResource which will set an environ on rendering"""
        return WSGIResource(self._reactor, self._threadpool,
                            self._application, environ)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2011 Manuel Stocker <mensi@mensi.ch>
#
# This file is part of GitServerGlue.
#
# GitServerGlue is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GitServerGlue is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GitServerGlue.  If not, see http://www.gnu.org/licenses

from setuptools import setup, find_packages

setup(name='GitServerGlue',
      install_requires=['twisted', 'pycrypto', 'passlib', 'pyasn1'],
      description=(
                   'Twisted-based implementation of the '
                   'network protocols supported by git'
                   ' (git://, ssh:// and http://)'
      ),
      keywords='git ssh http',
      version='0.3',
      obsoletes=['TwistedGit'],
      url='https://github.com/mensi/gitserverglue',
      license='GPL',
      author='Manuel Stocker',
      author_email='mensi@mensi.ch',
      long_description=(
                        'Twisted-based implementation of the '
                        'network protocols supported by git'
                        ' (git://, ssh:// and http://)'
                        ),
      packages=find_packages(),
      zip_safe=True,
      entry_points={'console_scripts': ['gitserverglue = gitserverglue:main']})



# -*- coding: utf-8 -*-
#
# Copyright 2011 Manuel Stocker <mensi@mensi.ch>
#
# This file is part of GitServerGlue.
#
# GitServerGlue is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GitServerGlue is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GitServerGlue.  If not, see http://www.gnu.org/licenses

from zope.interface import implements
from twisted.python import log

from twisted.internet import reactor
from twisted.internet.protocol import Protocol, ProcessProtocol, Factory
from twisted.internet.interfaces import IPushProducer

from gitserverglue.common import git_packet


class GitProcessProtocol(ProcessProtocol):
    def __init__(self, gitprotocol):
        self.gitprotocol = gitprotocol

    def connectionMade(self):
        # twisted.internet.process.Process seems to not fully
        # implement IPushProducer since stopProducing is missing
        # therefore patch in a dummy one
        if not hasattr(self.transport, "stopProducing"):
            setattr(self.transport, "stopProducing",
                    lambda: self.transport.loseConnection())

        self.transport.registerProducer(self.gitprotocol, True)
        self.gitprotocol.transport.registerProducer(self.transport, True)

        self.gitprotocol.resumeProducing()

    def outReceived(self, data):
        self.gitprotocol.transport.write(data)

    def errReceived(self, data):
        self.gitprotocol.transport.write(data)

    def processEnded(self, status):
        log.msg("Git ended with %r" % status)
        self.gitprotocol.transport.unregisterProducer()
        self.gitprotocol.transport.loseConnection()


class GitProtocol(Protocol):
    implements(IPushProducer)

    __buffer = ''
    paused = False
    requestReceived = False

    def __init__(self, authnz, git_configuration):
        self.authnz = authnz
        self.git_configuration = git_configuration

    def dataReceived(self, data):
        self.__buffer = self.__buffer + data

        while not self.paused and len(self.__buffer) >= 4:
            try:
                pktlen = int(self.__buffer[:4], 16)
            except ValueError:
                return self.sendErrorAndDisconnect(
                    "ERR Invalid Paket Length: " + self.__buffer[:4])

            if pktlen == 0:  # flush packet 0000
                pktlen = 4

            # The git protocol specifies bounds for the packet length
            if pktlen < 4 or pktlen > 65524:
                return self.sendErrorAndDisconnect(
                    "ERR Invalid Paket Length: " + self.__buffer[:4])

            # Do we have the complete packet in the buffer?
            if pktlen > len(self.__buffer):
                return

            packet = self.__buffer[:pktlen]
            self.__buffer = self.__buffer[pktlen:]
            self.packetReceived(packet)

    def packetReceived(self, data):
        if not self.requestReceived:
            payload = data[4:]

            # git:// would also support other RPC methods, but since
            # there is no authentication, only allow cloning aka
            # git-upload-pack
            if not payload.startswith("git-upload-pack"):
                return self.sendErrorAndDisconnect(
                    "ERR Request not supported. "
                    "Only git-upload-pack will be accepted")

            try:
                rpc_params = payload[len("git-upload-pack "):].split("\0")
                path, unused_host, unused_eol = rpc_params
            except ValueError:
                return self.sendErrorAndDisconnect(
                    "ERR Unable to parse request line")

            path_info = self.git_configuration.path_lookup(path,
                                                           protocol_hint='git')
            if path_info is None or path_info['repository_fs_path'] is None:
                return self.sendErrorAndDisconnect("ERR Repository not found")

            if not self.authnz.can_read(None, path_info):
                return self.sendErrorAndDisconnect(
                    "ERR Repository does not allow anonymous read access")

            # wait with data until we have a connection to the process
            self.pauseProducing()
            self.requestReceived = True
            self.process = GitProcessProtocol(self)

            gitbinary = self.git_configuration.git_binary
            cmdargs = ['git', 'upload-pack', path_info['repository_fs_path']]
            log.msg("Spawning %s with args %r" % (gitbinary, cmdargs))
            reactor.spawnProcess(self.process, gitbinary, cmdargs)

        else:
            self.process.transport.write(data)

    def sendErrorAndDisconnect(self, msg):
        self.transport.write(git_packet(msg))
        self.transport.loseConnection()

        # return None so it can be used
        # in a return statement in dataReceived for simplicity
        return None

    # IPushProducer
    def pauseProducing(self):
        self.paused = True
        self.transport.pauseProducing()

    def resumeProducing(self):
        self.paused = False
        self.transport.resumeProducing()
        self.dataReceived('')

    def stopProducing(self):
        # Only pause and don't call self.transport.stopProducing
        # since stopProducing will call loseConnection. This
        # can happen when git closes stdin but there is still
        # data on stdout/stderr
        # loseConnection will be called when the process ends
        # and everything has been written
        self.paused = True


class GitFactory(Factory):
    def __init__(self, authnz, git_configuration):
        self.authnz = authnz
        self.git_configuration = git_configuration

    def buildProtocol(self, addr):
        return GitProtocol(self.authnz, self.git_configuration)


def create_factory(authnz, git_configuration):
    return GitFactory(authnz, git_configuration)
